@@ -2,7 +2,7 @@
                           ftdi.c  -  description
                              -------------------
     begin                : Fri Apr 4 2003
-    copyright            : (C) 2003-2008 by Intra2net AG
+    copyright            : (C) 2003-2010 by Intra2net AG
     email                : opensource@intra2net.com
  ***************************************************************************/
 
@@ -64,7 +64,7 @@ static int ftdi_usb_close_internal (struct ftdi_context *ftdi)
 {
     int ret = 0;
 
-    if (ftdi->usb_dev)
+    if (ftdi && ftdi->usb_dev)
     {
        ret = usb_close (ftdi->usb_dev);
        ftdi->usb_dev = NULL;
@@ -159,9 +159,13 @@ struct ftdi_context *ftdi_new(void)
 
     \retval  0: all fine
     \retval -1: unknown interface
+    \retval -2: USB device unavailable
 */
 int ftdi_set_interface(struct ftdi_context *ftdi, enum ftdi_interface interface)
 {
+    if (ftdi == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     switch (interface)
     {
         case INTERFACE_ANY:
@@ -199,6 +203,9 @@ int ftdi_set_interface(struct ftdi_context *ftdi, enum ftdi_interface interface)
 */
 void ftdi_deinit(struct ftdi_context *ftdi)
 {
+    if (ftdi == NULL)
+        return;
+
     ftdi_usb_close_internal (ftdi);
 
     if (ftdi->async_usb_buffer != NULL)
@@ -233,6 +240,9 @@ void ftdi_free(struct ftdi_context *ftdi)
 */
 void ftdi_set_usbdev (struct ftdi_context *ftdi, usb_dev_handle *usb)
 {
+    if (ftdi == NULL)
+        return;
+
     ftdi->usb_dev = usb;
 }
 
@@ -396,6 +406,10 @@ static unsigned int _ftdi_determine_max_packet_size(struct ftdi_context *ftdi, s
 {
     unsigned int packet_size;
 
+    // Sanity check
+    if (ftdi == NULL || dev == NULL)
+        return 64;
+
     // Determine maximum packet size. Init with default value.
     // New hi-speed devices from FTDI use a packet size of 512 bytes
     // but could be connected to a normal speed USB hub -> 64 bytes packet size.
@@ -437,11 +451,16 @@ static unsigned int _ftdi_determine_max_packet_size(struct ftdi_context *ftdi, s
     \retval -5: unable to claim device
     \retval -6: reset failed
     \retval -7: set baudrate failed
+    \retval -8: ftdi context invalid
 */
 int ftdi_usb_open_dev(struct ftdi_context *ftdi, struct usb_device *dev)
 {
     int detach_errno = 0;
     int config_val = 1;
+
+    if (ftdi == NULL)
+        ftdi_error_return(-8, "ftdi context invalid");
+
     if (!(ftdi->usb_dev = usb_open(dev)))
         ftdi_error_return(-4, "usb_open() failed");
 
@@ -478,7 +497,7 @@ int ftdi_usb_open_dev(struct ftdi_context *ftdi, struct usb_device *dev)
             }
             else
             {
-                ftdi_error_return(-3, "unable to set usb configuration. Make sure ftdi_sio is unloaded!");
+                ftdi_error_return(-3, "unable to set usb configuration. Make sure the default FTDI driver is not in use");
             }
         }
     }
@@ -493,7 +512,7 @@ int ftdi_usb_open_dev(struct ftdi_context *ftdi, struct usb_device *dev)
         }
         else
         {
-            ftdi_error_return(-5, "unable to claim usb device. Make sure ftdi_sio is unloaded!");
+            ftdi_error_return(-5, "unable to claim usb device. Make sure the default FTDI driver is not in use");
         }
     }
 
@@ -608,6 +627,7 @@ int ftdi_usb_open_desc(struct ftdi_context *ftdi, int vendor, int product,
     \retval -8: get product description failed
     \retval -9: get serial number failed
     \retval -10: unable to close device
+    \retval -11: ftdi context invalid
 */
 int ftdi_usb_open_desc_index(struct ftdi_context *ftdi, int vendor, int product,
                        const char* description, const char* serial, unsigned int index)
@@ -623,6 +643,9 @@ int ftdi_usb_open_desc_index(struct ftdi_context *ftdi, int vendor, int product,
     if (usb_find_devices() < 0)
         ftdi_error_return(-2, "usb_find_devices() failed");
 
+    if (ftdi == NULL)
+        ftdi_error_return(-11, "ftdi context invalid");
+
     for (bus = usb_get_busses(); bus; bus = bus->next)
     {
         for (dev = bus->devices; dev; dev = dev->next)
@@ -705,9 +728,13 @@ int ftdi_usb_open_desc_index(struct ftdi_context *ftdi, int vendor, int product,
     \retval -9: get serial number failed
     \retval -10: unable to close device
     \retval -11: illegal description format
+    \retval -12: ftdi context invalid
 */
 int ftdi_usb_open_string(struct ftdi_context *ftdi, const char* description)
 {
+    if (ftdi == NULL)
+        ftdi_error_return(-12, "ftdi context invalid");
+
     if (description[0] == 0 || description[1] != ':')
         ftdi_error_return(-11, "illegal description format");
 
@@ -715,7 +742,6 @@ int ftdi_usb_open_string(struct ftdi_context *ftdi, const char* description)
     {
         struct usb_bus *bus;
         struct usb_device *dev;
-        char dev_name[PATH_MAX+1];
 
         usb_init();
 
@@ -728,9 +754,18 @@ int ftdi_usb_open_string(struct ftdi_context *ftdi, const char* description)
         {
             for (dev = bus->devices; dev; dev = dev->next)
             {
-                snprintf(dev_name, sizeof(dev_name), "%s/%s",bus->dirname,dev->filename);
-                if (strcmp(description+2,dev_name) == 0)
-                    return ftdi_usb_open_dev(ftdi, dev);
+                /* XXX: This doesn't handle symlinks/odd paths/etc... */
+                const char *desc = description + 2;
+                size_t len = strlen(bus->dirname);
+                if (strncmp(desc, bus->dirname, len))
+                    continue;
+                desc += len;
+                if (desc[0] != '/')
+                    continue;
+                ++desc;
+                if (strcmp(desc, dev->filename))
+                    continue;
+                return ftdi_usb_open_dev(ftdi, dev);
             }
         }
 
@@ -791,9 +826,13 @@ int ftdi_usb_open_string(struct ftdi_context *ftdi, const char* description)
 
     \retval  0: all fine
     \retval -1: FTDI reset failed
+    \retval -2: USB device unavailable
 */
 int ftdi_usb_reset(struct ftdi_context *ftdi)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE,
                         SIO_RESET_REQUEST, SIO_RESET_SIO,
                         ftdi->index, NULL, 0, ftdi->usb_write_timeout) != 0)
@@ -813,9 +852,13 @@ int ftdi_usb_reset(struct ftdi_context *ftdi)
 
     \retval  0: all fine
     \retval -1: read buffer purge failed
+    \retval -2: USB device unavailable
 */
 int ftdi_usb_purge_rx_buffer(struct ftdi_context *ftdi)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE,
                         SIO_RESET_REQUEST, SIO_RESET_PURGE_RX,
                         ftdi->index, NULL, 0, ftdi->usb_write_timeout) != 0)
@@ -835,9 +878,13 @@ int ftdi_usb_purge_rx_buffer(struct ftdi_context *ftdi)
 
     \retval  0: all fine
     \retval -1: write buffer purge failed
+    \retval -2: USB device unavailable
 */
 int ftdi_usb_purge_tx_buffer(struct ftdi_context *ftdi)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE,
                         SIO_RESET_REQUEST, SIO_RESET_PURGE_TX,
                         ftdi->index, NULL, 0, ftdi->usb_write_timeout) != 0)
@@ -854,11 +901,15 @@ int ftdi_usb_purge_tx_buffer(struct ftdi_context *ftdi)
     \retval  0: all fine
     \retval -1: read buffer purge failed
     \retval -2: write buffer purge failed
+    \retval -3: USB device unavailable
 */
 int ftdi_usb_purge_buffers(struct ftdi_context *ftdi)
 {
     int result;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-3, "USB device unavailable");
+
     result = ftdi_usb_purge_rx_buffer(ftdi);
     if (result < 0)
         return -1;
@@ -880,11 +931,15 @@ int ftdi_usb_purge_buffers(struct ftdi_context *ftdi)
     \retval  0: all fine
     \retval -1: usb_release failed
     \retval -2: usb_close failed
+    \retval -3: ftdi context invalid
 */
 int ftdi_usb_close(struct ftdi_context *ftdi)
 {
     int rtn = 0;
 
+    if (ftdi == NULL)
+        ftdi_error_return(-3, "ftdi context invalid");
+
 #ifdef LIBFTDI_LINUX_ASYNC_MODE
     /* try to release some kernel resources */
     ftdi_async_complete(ftdi,1);
@@ -1035,12 +1090,16 @@ static int ftdi_convert_baudrate(int baudrate, struct ftdi_context *ftdi,
     \retval  0: all fine
     \retval -1: invalid baudrate
     \retval -2: setting baudrate failed
+    \retval -3: USB device unavailable
 */
 int ftdi_set_baudrate(struct ftdi_context *ftdi, int baudrate)
 {
     unsigned short value, index;
     int actual_baudrate;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-3, "USB device unavailable");
+
     if (ftdi->bitbang_enabled)
     {
         baudrate = baudrate*4;
@@ -1096,6 +1155,7 @@ int ftdi_set_line_property(struct ftdi_context *ftdi, enum ftdi_bits_type bits,
 
     \retval  0: all fine
     \retval -1: Setting line property failed
+    \retval -2: USB device unavailable
 */
 int ftdi_set_line_property2(struct ftdi_context *ftdi, enum ftdi_bits_type bits,
                             enum ftdi_stopbits_type sbit, enum ftdi_parity_type parity,
@@ -1103,6 +1163,9 @@ int ftdi_set_line_property2(struct ftdi_context *ftdi, enum ftdi_bits_type bits,
 {
     unsigned short value = bits;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     switch (parity)
     {
         case NONE:
@@ -1160,6 +1223,7 @@ int ftdi_set_line_property2(struct ftdi_context *ftdi, enum ftdi_bits_type bits,
     \param buf Buffer with the data
     \param size Size of the buffer
 
+    \retval -666: USB device unavailable
     \retval <0: error code from usb_bulk_write()
     \retval >0: number of bytes written
 */
@@ -1169,6 +1233,9 @@ int ftdi_write_data(struct ftdi_context *ftdi, unsigned char *buf, int size)
     int offset = 0;
     int total_written = 0;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-666, "USB device unavailable");
+
     while (offset < size)
     {
         int write_size = ftdi->writebuffer_chunksize;
@@ -1373,6 +1440,7 @@ static int _usb_bulk_write_async(struct ftdi_context *ftdi, int ep, char *bytes,
     \param buf Buffer with the data
     \param size Size of the buffer
 
+    \retval -666: USB device unavailable
     \retval <0: error code from usb_bulk_write()
     \retval >0: number of bytes written
 */
@@ -1382,6 +1450,9 @@ int ftdi_write_data_async(struct ftdi_context *ftdi, unsigned char *buf, int siz
     int offset = 0;
     int total_written = 0;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-666, "USB device unavailable");
+
     while (offset < size)
     {
         int write_size = ftdi->writebuffer_chunksize;
@@ -1409,9 +1480,13 @@ int ftdi_write_data_async(struct ftdi_context *ftdi, unsigned char *buf, int siz
     \param chunksize Chunk size
 
     \retval 0: all fine
+    \retval -1: ftdi context invalid
 */
 int ftdi_write_data_set_chunksize(struct ftdi_context *ftdi, unsigned int chunksize)
 {
+    if (ftdi == NULL)
+        ftdi_error_return(-1, "ftdi context invalid");
+
     ftdi->writebuffer_chunksize = chunksize;
     return 0;
 }
@@ -1423,9 +1498,13 @@ int ftdi_write_data_set_chunksize(struct ftdi_context *ftdi, unsigned int chunks
     \param chunksize Pointer to store chunk size in
 
     \retval 0: all fine
+    \retval -1: ftdi context invalid
 */
 int ftdi_write_data_get_chunksize(struct ftdi_context *ftdi, unsigned int *chunksize)
 {
+    if (ftdi == NULL)
+        ftdi_error_return(-1, "ftdi context invalid");
+
     *chunksize = ftdi->writebuffer_chunksize;
     return 0;
 }
@@ -1439,6 +1518,7 @@ int ftdi_write_data_get_chunksize(struct ftdi_context *ftdi, unsigned int *chunk
     \param buf Buffer to store data in
     \param size Size of the buffer
 
+    \retval -666: USB device unavailable
     \retval <0: error code from usb_bulk_read()
     \retval  0: no data was available
     \retval >0: number of bytes read
@@ -1447,8 +1527,12 @@ int ftdi_write_data_get_chunksize(struct ftdi_context *ftdi, unsigned int *chunk
 int ftdi_read_data(struct ftdi_context *ftdi, unsigned char *buf, int size)
 {
     int offset = 0, ret = 1, i, num_of_chunks, chunk_remains;
-    int packet_size = ftdi->max_packet_size;
+    int packet_size;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-666, "USB device unavailable");
+
+    packet_size = ftdi->max_packet_size;
     // Packet size sanity check (avoid division by zero)
     if (packet_size == 0)
         ftdi_error_return(-1, "max_packet_size is bogus (zero)");
@@ -1563,11 +1647,15 @@ int ftdi_read_data(struct ftdi_context *ftdi, unsigned char *buf, int size)
     \param chunksize Chunk size
 
     \retval 0: all fine
+    \retval -1: ftdi context invalid
 */
 int ftdi_read_data_set_chunksize(struct ftdi_context *ftdi, unsigned int chunksize)
 {
     unsigned char *new_buf;
 
+    if (ftdi == NULL)
+        ftdi_error_return(-1, "ftdi context invalid");
+
     // Invalidate all remaining data
     ftdi->readbuffer_offset = 0;
     ftdi->readbuffer_remaining = 0;
@@ -1588,9 +1676,13 @@ int ftdi_read_data_set_chunksize(struct ftdi_context *ftdi, unsigned int chunksi
     \param chunksize Pointer to store chunk size in
 
     \retval 0: all fine
+    \retval -1: FTDI context invalid
 */
 int ftdi_read_data_get_chunksize(struct ftdi_context *ftdi, unsigned int *chunksize)
 {
+    if (ftdi == NULL)
+        ftdi_error_return(-1, "FTDI context invalid");
+
     *chunksize = ftdi->readbuffer_chunksize;
     return 0;
 }
@@ -1607,11 +1699,15 @@ int ftdi_read_data_get_chunksize(struct ftdi_context *ftdi, unsigned int *chunks
 
     \retval  0: all fine
     \retval -1: can't enable bitbang mode
+    \retval -2: USB device unavailable
 */
 int ftdi_enable_bitbang(struct ftdi_context *ftdi, unsigned char bitmask)
 {
     unsigned short usb_val;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     usb_val = bitmask; // low byte: bitmask
     /* FT2232C: Set bitbang_mode to 2 to enable SPI */
     usb_val |= (ftdi->bitbang_mode << 8);
@@ -1632,9 +1728,13 @@ int ftdi_enable_bitbang(struct ftdi_context *ftdi, unsigned char bitmask)
 
     \retval  0: all fine
     \retval -1: can't disable bitbang mode
+    \retval -2: USB device unavailable
 */
 int ftdi_disable_bitbang(struct ftdi_context *ftdi)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE, SIO_SET_BITMODE_REQUEST, 0, ftdi->index, NULL, 0, ftdi->usb_write_timeout) != 0)
         ftdi_error_return(-1, "unable to leave bitbang mode. Perhaps not a BM type chip?");
 
@@ -1652,11 +1752,15 @@ int ftdi_disable_bitbang(struct ftdi_context *ftdi)
 
     \retval  0: all fine
     \retval -1: can't enable bitbang mode
+    \retval -2: USB device unavailable
 */
 int ftdi_set_bitmode(struct ftdi_context *ftdi, unsigned char bitmask, unsigned char mode)
 {
     unsigned short usb_val;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     usb_val = bitmask; // low byte: bitmask
     usb_val |= (mode << 8);
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE, SIO_SET_BITMODE_REQUEST, usb_val, ftdi->index, NULL, 0, ftdi->usb_write_timeout) != 0)
@@ -1675,9 +1779,13 @@ int ftdi_set_bitmode(struct ftdi_context *ftdi, unsigned char bitmask, unsigned
 
     \retval  0: all fine
     \retval -1: read pins failed
+    \retval -2: USB device unavailable
 */
 int ftdi_read_pins(struct ftdi_context *ftdi, unsigned char *pins)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_IN_REQTYPE, SIO_READ_PINS_REQUEST, 0, ftdi->index, (char *)pins, 1, ftdi->usb_read_timeout) != 1)
         ftdi_error_return(-1, "read pins failed");
 
@@ -1697,6 +1805,7 @@ int ftdi_read_pins(struct ftdi_context *ftdi, unsigned char *pins)
     \retval  0: all fine
     \retval -1: latency out of range
     \retval -2: unable to set latency timer
+    \retval -3: USB device unavailable
 */
 int ftdi_set_latency_timer(struct ftdi_context *ftdi, unsigned char latency)
 {
@@ -1705,6 +1814,9 @@ int ftdi_set_latency_timer(struct ftdi_context *ftdi, unsigned char latency)
     if (latency < 1)
         ftdi_error_return(-1, "latency out of range. Only valid for 1-255");
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-3, "USB device unavailable");
+
     usb_val = latency;
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE, SIO_SET_LATENCY_TIMER_REQUEST, usb_val, ftdi->index, NULL, 0, ftdi->usb_write_timeout) != 0)
         ftdi_error_return(-2, "unable to set latency timer");
@@ -1720,10 +1832,15 @@ int ftdi_set_latency_timer(struct ftdi_context *ftdi, unsigned char latency)
 
     \retval  0: all fine
     \retval -1: unable to get latency timer
+    \retval -2: USB device unavailable
 */
 int ftdi_get_latency_timer(struct ftdi_context *ftdi, unsigned char *latency)
 {
     unsigned short usb_val;
+
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_IN_REQTYPE, SIO_GET_LATENCY_TIMER_REQUEST, 0, ftdi->index, (char *)&usb_val, 1, ftdi->usb_read_timeout) != 1)
         ftdi_error_return(-1, "reading latency timer failed");
 
@@ -1769,11 +1886,15 @@ int ftdi_get_latency_timer(struct ftdi_context *ftdi, unsigned char *latency)
 
     \retval  0: all fine
     \retval -1: unable to retrieve status information
+    \retval -2: USB device unavailable
 */
 int ftdi_poll_modem_status(struct ftdi_context *ftdi, unsigned short *status)
 {
     char usb_val[2];
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_IN_REQTYPE, SIO_POLL_MODEM_STATUS_REQUEST, 0, ftdi->index, usb_val, 2, ftdi->usb_read_timeout) != 2)
         ftdi_error_return(-1, "getting modem status failed");
 
@@ -1791,9 +1912,13 @@ int ftdi_poll_modem_status(struct ftdi_context *ftdi, unsigned short *status)
 
     \retval  0: all fine
     \retval -1: set flow control failed
+    \retval -2: USB device unavailable
 */
 int ftdi_setflowctrl(struct ftdi_context *ftdi, int flowctrl)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE,
                         SIO_SET_FLOW_CTRL_REQUEST, 0, (flowctrl | ftdi->index),
                         NULL, 0, ftdi->usb_write_timeout) != 0)
@@ -1810,11 +1935,15 @@ int ftdi_setflowctrl(struct ftdi_context *ftdi, int flowctrl)
 
     \retval  0: all fine
     \retval -1: set dtr failed
+    \retval -2: USB device unavailable
 */
 int ftdi_setdtr(struct ftdi_context *ftdi, int state)
 {
     unsigned short usb_val;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (state)
         usb_val = SIO_SET_DTR_HIGH;
     else
@@ -1835,12 +1964,16 @@ int ftdi_setdtr(struct ftdi_context *ftdi, int state)
     \param state state to set line to (1 or 0)
 
     \retval  0: all fine
-    \retval -1 set rts failed
+    \retval -1: set rts failed
+    \retval -2: USB device unavailable
 */
 int ftdi_setrts(struct ftdi_context *ftdi, int state)
 {
     unsigned short usb_val;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (state)
         usb_val = SIO_SET_RTS_HIGH;
     else
@@ -1855,19 +1988,23 @@ int ftdi_setrts(struct ftdi_context *ftdi, int state)
 }
 
 /**
- Set dtr and rts line in one pass
+    Set dtr and rts line in one pass
 
- \param ftdi pointer to ftdi_context
- \param dtr  DTR state to set line to (1 or 0)
- \param rts  RTS state to set line to (1 or 0)
+    \param ftdi pointer to ftdi_context
+    \param dtr  DTR state to set line to (1 or 0)
+    \param rts  RTS state to set line to (1 or 0)
 
- \retval  0: all fine
- \retval -1 set dtr/rts failed
+    \retval  0: all fine
+    \retval -1: set dtr/rts failed
+    \retval -2: USB device unavailable
  */
 int ftdi_setdtr_rts(struct ftdi_context *ftdi, int dtr, int rts)
 {
     unsigned short usb_val;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (dtr)
         usb_val = SIO_SET_DTR_HIGH;
     else
@@ -1895,12 +2032,16 @@ int ftdi_setdtr_rts(struct ftdi_context *ftdi, int dtr, int rts)
 
     \retval  0: all fine
     \retval -1: unable to set event character
+    \retval -2: USB device unavailable
 */
 int ftdi_set_event_char(struct ftdi_context *ftdi,
                         unsigned char eventch, unsigned char enable)
 {
     unsigned short usb_val;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     usb_val = eventch;
     if (enable)
         usb_val |= 1 << 8;
@@ -1920,12 +2061,16 @@ int ftdi_set_event_char(struct ftdi_context *ftdi,
 
     \retval  0: all fine
     \retval -1: unable to set error character
+    \retval -2: USB device unavailable
 */
 int ftdi_set_error_char(struct ftdi_context *ftdi,
                         unsigned char errorch, unsigned char enable)
 {
     unsigned short usb_val;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     usb_val = errorch;
     if (enable)
         usb_val |= 1 << 8;
@@ -1946,6 +2091,9 @@ int ftdi_set_error_char(struct ftdi_context *ftdi,
 */
 void ftdi_eeprom_setsize(struct ftdi_context *ftdi, struct ftdi_eeprom *eeprom, int size)
 {
+    if (ftdi == NULL)
+        return;
+
     ftdi->eeprom_size=size;
     eeprom->size=size;
 }
@@ -1957,6 +2105,9 @@ void ftdi_eeprom_setsize(struct ftdi_context *ftdi, struct ftdi_eeprom *eeprom,
 */
 void ftdi_eeprom_initdefaults(struct ftdi_eeprom *eeprom)
 {
+    if (eeprom == NULL)
+        return;
+
     eeprom->vendor_id = 0x0403;
     eeprom->product_id = 0x6001;
 
@@ -1981,14 +2132,39 @@ void ftdi_eeprom_initdefaults(struct ftdi_eeprom *eeprom)
 }
 
 /**
-   Build binary output from ftdi_eeprom structure.
-   Output is suitable for ftdi_write_eeprom().
+    Frees allocated memory in eeprom.
 
-   \param eeprom Pointer to ftdi_eeprom
-   \param output Buffer of 128 bytes to store eeprom image to
+    \param eeprom Pointer to ftdi_eeprom
+*/
+void ftdi_eeprom_free(struct ftdi_eeprom *eeprom)
+{
+    if (!eeprom)
+        return;
 
-   \retval >0: used eeprom size
-   \retval -1: eeprom size (128 bytes) exceeded by custom strings
+    if (eeprom->manufacturer != 0) {
+        free(eeprom->manufacturer);
+        eeprom->manufacturer = 0;
+    }
+    if (eeprom->product != 0) {
+        free(eeprom->product);
+        eeprom->product = 0;
+    }
+    if (eeprom->serial != 0) {
+        free(eeprom->serial);
+        eeprom->serial = 0;
+    }
+}
+
+/**
+    Build binary output from ftdi_eeprom structure.
+    Output is suitable for ftdi_write_eeprom().
+
+    \param eeprom Pointer to ftdi_eeprom
+    \param output Buffer of 128 bytes to store eeprom image to
+
+    \retval >0: used eeprom size
+    \retval -1: eeprom size (128 bytes) exceeded by custom strings
+    \retval -2: Invalid eeprom pointer
 */
 int ftdi_eeprom_build(struct ftdi_eeprom *eeprom, unsigned char *output)
 {
@@ -1997,6 +2173,9 @@ int ftdi_eeprom_build(struct ftdi_eeprom *eeprom, unsigned char *output)
     unsigned char manufacturer_size = 0, product_size = 0, serial_size = 0;
     int size_check;
 
+    if (eeprom == NULL)
+        return -2;
+
     if (eeprom->manufacturer != NULL)
         manufacturer_size = strlen(eeprom->manufacturer);
     if (eeprom->product != NULL)
@@ -2175,6 +2354,9 @@ int ftdi_eeprom_decode(struct ftdi_eeprom *eeprom, unsigned char *buf, int size)
     unsigned char manufacturer_size = 0, product_size = 0, serial_size = 0;
     int size_check;
     int eeprom_size = 128;
+
+    if (eeprom == NULL)
+        return -1;
 #if 0
     size_check = eeprom->size;
     size_check -= 28; // 28 are always in use (fixed)
@@ -2330,9 +2512,13 @@ int ftdi_eeprom_decode(struct ftdi_eeprom *eeprom, unsigned char *buf, int size)
 
     \retval  0: all fine
     \retval -1: read failed
+    \retval -2: USB device unavailable
 */
 int ftdi_read_eeprom_location (struct ftdi_context *ftdi, int eeprom_addr, unsigned short *eeprom_val)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_IN_REQTYPE, SIO_READ_EEPROM_REQUEST, 0, eeprom_addr, (char *)eeprom_val, 2, ftdi->usb_read_timeout) != 2)
         ftdi_error_return(-1, "reading eeprom failed");
 
@@ -2347,11 +2533,15 @@ int ftdi_read_eeprom_location (struct ftdi_context *ftdi, int eeprom_addr, unsig
 
     \retval  0: all fine
     \retval -1: read failed
+    \retval -2: USB device unavailable
 */
 int ftdi_read_eeprom(struct ftdi_context *ftdi, unsigned char *eeprom)
 {
     int i;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     for (i = 0; i < ftdi->eeprom_size/2; i++)
     {
         if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_IN_REQTYPE, SIO_READ_EEPROM_REQUEST, 0, i, eeprom+(i*2), 2, ftdi->usb_read_timeout) != 2)
@@ -2386,11 +2576,15 @@ static unsigned char ftdi_read_chipid_shift(unsigned char value)
 
     \retval  0: all fine
     \retval -1: read failed
+    \retval -2: USB device unavailable
 */
 int ftdi_read_chipid(struct ftdi_context *ftdi, unsigned int *chipid)
 {
     unsigned int a = 0, b = 0;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_IN_REQTYPE, SIO_READ_EEPROM_REQUEST, 0, 0x43, (char *)&a, 2, ftdi->usb_read_timeout) == 2)
     {
         a = a << 8 | a >> 8;
@@ -2409,20 +2603,25 @@ int ftdi_read_chipid(struct ftdi_context *ftdi, unsigned int *chipid)
 }
 
 /**
-   Guesses size of eeprom by reading eeprom and comparing halves - will not work with blank eeprom
-   Call this function then do a write then call again to see if size changes, if so write again.
+    Guesses size of eeprom by reading eeprom and comparing halves - will not work with blank eeprom
+    Call this function then do a write then call again to see if size changes, if so write again.
 
-   \param ftdi pointer to ftdi_context
-   \param eeprom Pointer to store eeprom into
-   \param maxsize the size of the buffer to read into
+    \param ftdi pointer to ftdi_context
+    \param eeprom Pointer to store eeprom into
+    \param maxsize the size of the buffer to read into
 
-   \retval size of eeprom
+    \retval -1: eeprom read failed
+    \retval -2: USB device unavailable
+    \retval >=0: size of eeprom
 */
 int ftdi_read_eeprom_getsize(struct ftdi_context *ftdi, unsigned char *eeprom, int maxsize)
 {
     int i=0,j,minsize=32;
     int size=minsize;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     do
     {
         for (j = 0; i < maxsize/2 && j<size; j++)
@@ -2430,7 +2629,7 @@ int ftdi_read_eeprom_getsize(struct ftdi_context *ftdi, unsigned char *eeprom, i
             if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_IN_REQTYPE,
                                 SIO_READ_EEPROM_REQUEST, 0, i,
                                 eeprom+(i*2), 2, ftdi->usb_read_timeout) != 2)
-                ftdi_error_return(-1, "reading eeprom failed");
+                ftdi_error_return(-1, "eeprom read failed");
             i++;
         }
         size*=2;
@@ -2449,9 +2648,13 @@ int ftdi_read_eeprom_getsize(struct ftdi_context *ftdi, unsigned char *eeprom, i
 
     \retval  0: all fine
     \retval -1: read failed
+    \retval -2: USB device unavailable
 */
 int ftdi_write_eeprom_location(struct ftdi_context *ftdi, int eeprom_addr, unsigned short eeprom_val)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE,
                                     SIO_WRITE_EEPROM_REQUEST, eeprom_val, eeprom_addr,
                                     NULL, 0, ftdi->usb_write_timeout) != 0)
@@ -2468,12 +2671,16 @@ int ftdi_write_eeprom_location(struct ftdi_context *ftdi, int eeprom_addr, unsig
 
     \retval  0: all fine
     \retval -1: read failed
+    \retval -2: USB device unavailable
 */
 int ftdi_write_eeprom(struct ftdi_context *ftdi, unsigned char *eeprom)
 {
     unsigned short usb_val, status;
     int i, ret;
 
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     /* These commands were traced while running MProg */
     if ((ret = ftdi_usb_reset(ftdi)) != 0)
         return ret;
@@ -2504,9 +2711,13 @@ int ftdi_write_eeprom(struct ftdi_context *ftdi, unsigned char *eeprom)
 
     \retval  0: all fine
     \retval -1: erase failed
+    \retval -2: USB device unavailable
 */
 int ftdi_erase_eeprom(struct ftdi_context *ftdi)
 {
+    if (ftdi == NULL || ftdi->usb_dev == NULL)
+        ftdi_error_return(-2, "USB device unavailable");
+
     if (usb_control_msg(ftdi->usb_dev, FTDI_DEVICE_OUT_REQTYPE, SIO_ERASE_EEPROM_REQUEST, 0, 0, NULL, 0, ftdi->usb_write_timeout) != 0)
         ftdi_error_return(-1, "unable to erase eeprom");
 
@@ -2522,6 +2733,9 @@ int ftdi_erase_eeprom(struct ftdi_context *ftdi)
 */
 char *ftdi_get_error_string (struct ftdi_context *ftdi)
 {
+    if (ftdi == NULL)
+        return "";
+
     return ftdi->error_str;
 }
 
